import click
import os
import pyperclip
import subprocess
import tomllib

from .tools.preprocessor import preprocess

@click.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("--solution", "solution_src", default="solution", help="Which solution file to submit.")
@click.option("--open", "open_editor", is_flag=True, help="Open submission file on text editor.")
@click.option("--copy", is_flag=True, help="Copy submission to clipboard.")
def submit(**kwargs):
    """Bundle solution into a single source file for submission."""

    handle_submit(**kwargs)

def handle_submit(judge, problem, solution_src, open_editor, copy):
    with open(os.path.join(os.getcwd(), "icancrc.toml"), "rb") as f:
        cfg = tomllib.load(f)
    
    solution_dir = os.path.join(os.getcwd(), "problems", judge, problem)
    solution_path = os.path.join(solution_dir, f"{solution_src}.c")
    if not os.path.exists(solution_path):
        click.echo("Problem {}/{} does not exist.".format(judge, problem), err=True)
        exit(1)

    submission = preprocess(solution_path)

    submission_dir = os.path.join(os.getcwd(), "submissions", judge, problem)
    submission_path = os.path.join(submission_dir, f"{solution_src}.c")
    os.makedirs(submission_dir, exist_ok=True)

    with open(submission_path, "w") as f:
        f.write(submission)

    if open_editor:
        subprocess.run([cfg["editor"], submission_path])

    if copy:
        pyperclip.copy(submission)



import click
import os

def preprocess(path):
    header = "/*\n* Generated by icc\n* https://github.com/voxelstack/icc\n*/\n"
    return "{}\n{}".format(header, "".join(preprocess_source(path)["source"]))

def preprocess_source(path, cache={}, included=[]):
    if path in cache.values():
        return cache[path]
    
    with open(path, "r") as f:
        source = iter(f.readlines())
        preprocessed_source = []
        pragma_once = False
        last_line_empty = False

        line = next(source, None)
        while line is not None:
            trimmed = line.lstrip()

            if trimmed.startswith("#pragma"):
                pragma = trimmed[len("#pragma"):].lstrip()
                if pragma.startswith("once"):
                    pragma_once = True
                elif pragma.startswith("icc"):
                    pragma_icc = pragma[len("icc"):].lstrip()
                    if pragma_icc.startswith("include"):
                            while True:
                                line = next(source, None).lstrip()
                                if line is None or not line.startswith("#include"):
                                    if line.startswith("#pragma icc end"):
                                        break
                                    
                                    click.echo("Invalid icc include block. Expecting #include or #pragma icc end.", err=True)
                                    exit(1)

                                include_file = line[len("#include"):].strip()[1:-1]
                                include_path = os.path.join(os.getcwd(), "include", include_file)
                                include = preprocess_source(include_path, cache, included)
                                if not include["once"] or include_path not in included:
                                    preprocessed_source.append("// BEGIN {}\n".format(include_file))
                                    preprocessed_source.extend(include["source"])
                                    preprocessed_source.append("// END {}\n".format(include_file))
                                    
                                    included.append(include_path)
                    else:
                        click.echo("Unknown pragma {}.".format(line))
                line = next(source, None)
                continue

            current_line_empty = len(trimmed) == 0
            # Since we are ommiting some lines when generating the submission,
            # it's common to have double line breaks which look bad.
            # Skip them.
            if not (last_line_empty and current_line_empty):
                preprocessed_source.append(line)
            last_line_empty = current_line_empty

            line = next(source, None)

    source = {
        "source": preprocessed_source,
        "once": pragma_once,
    }
    cache[path] = source
    
    return source



import click
import os
import shutil
import subprocess

@click.command()
@click.option("--name", prompt=True, default="leet")
def init(name):
    """Initialize an empty icanc project."""

    dir = os.path.join(os.getcwd(), name)
    if os.path.exists(dir):
        click.echo("Directory ./{} already exists.".format(name), err=True)
        exit(1)
    os.makedirs(dir)

    res_dir = os.path.join(os.path.dirname(__file__), "..", "data")

    os.makedirs(os.path.join(dir, "problems"))
    shutil.copy2(os.path.join(res_dir, "icancrc.toml"), dir)
    shutil.copy2(os.path.join(res_dir, "LICENSE"), dir)
    shutil.copytree(os.path.join(res_dir, "include"), os.path.join(dir, "include"))
    shutil.copytree(os.path.join(res_dir, "templates"), os.path.join(dir, "templates"))
    with open(os.path.join(res_dir, "README.md"), "r") as src:
        readme = src.read().format(name=name)
        with open(os.path.join(dir, "README.md"), "w") as dst:
            dst.write(readme)
    
    click.echo("\nInitialized icc project at ./{}".format(name))
    click.echo("Get started with cd ./{} && icc --help".format(name))

def read_git_config(config):
    res = subprocess.run(["git", "config", config], stdout=subprocess.PIPE)
    if res.returncode == 0:
        return res.stdout.rstrip().decode()
    return ""



import click
import os
import shutil
import subprocess
import tomllib

@click.group()
def create():
    """Create solutions or testcases."""

    pass

@create.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("--template", type=click.Path(exists=True), required=True)
@click.option("--solution", "solution_dst", default="solution", help="Name for the solution file.")
@click.option("--open", "open_editor", is_flag=True, help="Open solution file on text editor.")
def solution(**kwargs):
    create_solution(**kwargs)

def create_solution(judge, problem, template, solution_dst, open_editor):
    with open(os.path.join(os.getcwd(), "icancrc.toml"), "rb") as f:
        cfg = tomllib.load(f)
    
    dir = os.path.join(os.getcwd(), "problems", judge, problem)
    os.makedirs(dir, exist_ok=True)

    solution_path = os.path.join(dir, "{}.c".format(solution_dst))
    shutil.copy2(template, solution_path)
    
    click.echo("Created blank solution: ./problems/{}/{}/{}.c".format(judge, problem, solution_dst))

    if open_editor:
        subprocess.run([cfg["editor"], solution_path])

@create.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("--open", "open_editor", is_flag=True, help="Open testcases file on text editor.")
@click.option("--testcases", "testcases_dst", default="testcases", help="Name for the testcases file.")
def testcases(**kwargs):
    create_testcases(**kwargs)

def create_testcases(judge, problem, testcases_dst, open_editor):
    with open(os.path.join(os.getcwd(), "icancrc.toml"), "rb") as f:
        cfg = tomllib.load(f)
    
    dir = os.path.join(os.getcwd(), "problems", judge, problem)
    if not os.path.exists(dir):
        click.echo("Problem {}/{} does not exist.".format(judge, problem), err=True)
        exit(1)
    os.makedirs(dir, exist_ok=True)

    testcases_path = os.path.join(dir, "{}.toml".format(testcases_dst))
    
    with open(testcases_path, "w") as f:
        f.writelines([
            "# {}/{}\n\n".format(judge, problem),
            "[case1]\n",
            "in = \"\"\"\"\"\"\n",
            "out = \"\"\"\"\"\"\n"
        ])

    click.echo("Created blank testcases: ./problems/{}/{}/{}.toml".format(judge, problem, testcases_dst))

    if open_editor:
        subprocess.run([cfg["editor"], testcases_path])

@click.command()
@click.argument("judge", type=str)
@click.argument("problem", type=str)
@click.option("--template", type=click.Path(exists=True), required=True)
@click.option("--solution", "solution_dst", default="solution", help="Name for the solution file.")
@click.option("--testcases", "testcases_dst", default="testcases", help="Name for the testcases file.")
@click.option("--open", "open_editor", is_flag=True, help="Open solution file on text editor.")
def scaffold(judge, problem, template, solution_dst, testcases_dst, open_editor):
    """Create a solution and testcase files."""

    click.echo("Scaffolding {}/{}\n".format(judge, problem))

    create_solution(judge, problem, template, solution_dst, open_editor)
    create_testcases(judge, problem, testcases_dst, False)



import click

from .commands.create import create, scaffold
from .commands.init import init
from .commands.submit import submit
from .commands.test import test
from .commands.ci import ci

@click.group()
def icanc():
    pass

icanc.add_command(init)
icanc.add_command(create)
icanc.add_command(scaffold)
icanc.add_command(submit)
icanc.add_command(test)
icanc.add_command(ci)

def main():
    icanc()

if __name__ == '__main__':
    main()
